# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = ["Context"]

import datetime
import functools
import re
import typing

import hikari

if typing.TYPE_CHECKING:
    from lightbulb import command_handler
    from lightbulb import commands
    from lightbulb import plugins


class Context:
    """
    The context a command was invoked under.

    Args:
        bot (:obj:`~.command_handler.Bot`): The bot instance that received the command.
        message (:obj:`hikari.messages.Message`): The message the context was created from.
        prefix (:obj:`str`): The prefix used in the context.
        invoked_with (:obj:`str`): The name or alias used to invoke a command.
        command (:obj:`~.commands.Command`): The command that was invoked.

    Note:
        For information on types for the various properties see :obj:`hikari.messages.Message`.
    """

    __slots__: typing.Sequence[str] = ("_bot", "_message", "_prefix", "_invoked_with", "_command")

    def __init__(
        self,
        bot: command_handler.Bot,
        message: hikari.Message,
        prefix: str,
        invoked_with: str,
        command: commands.Command,
    ) -> None:
        self._bot: command_handler.Bot = bot
        """The bot instance."""
        self._message: hikari.Message = message
        """The message that the context derived from."""
        self._prefix: str = prefix
        """The command prefix used."""
        self._invoked_with: str = invoked_with
        """The command name or alias used."""
        self._command: commands.Command = command
        """The command that was invoked."""

    # XXX: should this be deprecated and changed to `app` to be consistent
    # with the entire hikari API, and the lightbulb error event API?
    @property
    def bot(self) -> command_handler.Bot:
        return self._bot

    @property
    def message(self) -> hikari.Message:
        return self._message

    @property
    def prefix(self) -> str:
        return self._prefix

    @property
    def invoked_with(self) -> str:
        return self._invoked_with

    @property
    def command(self) -> commands.Command:
        return self._command

    @property
    def guild_id(self) -> typing.Optional[hikari.Snowflake]:
        """ID of the guild the command was invoked in, or None if the command was invoked in DMs."""
        return self._message.guild_id

    @property
    def channel_id(self) -> hikari.Snowflake:
        """ID of the channel the command was invoked in."""
        return self._message.channel_id

    @property
    def content(self) -> str:
        """Raw content of the invocation message."""
        return self._message.content

    @property
    def member(self) -> typing.Optional[hikari.Member]:
        """Optional member corresponding to the context author."""
        return self._message.member

    @property
    def message_id(self) -> hikari.Snowflake:
        """ID of the message that invoked the command."""
        return self._message.id

    @property
    def timestamp(self) -> datetime.datetime:
        """The timestamp the context message was sent at."""
        return self._message.timestamp

    @property
    def edited_timestamp(self) -> typing.Optional[datetime.datetime]:
        """Optional timestamp of the previous edit of the context message."""
        return self._message.edited_timestamp

    # XXX: these are at the time of writing typing.Sequence objects, but will
    # be refactored to use typing.AbstractSet as part of https://github.com/nekokatt/hikari/issues/273.
    @property
    def user_mentions(self) -> typing.Collection[hikari.Snowflake]:
        """The users mentioned in the context message."""
        return self._message.user_mentions

    @property
    def role_mentions(self) -> typing.Collection[hikari.Snowflake]:
        """The roles mentioned in the context message."""
        return self._message.role_mentions

    @property
    def channel_mentions(self) -> typing.Collection[hikari.Snowflake]:
        """The channels mentioned in the context message."""
        return self._message.channel_mentions

    @property
    def attachments(self) -> typing.Sequence[hikari.Attachment]:
        """The attachments to the context message."""
        return self._message.attachments

    @property
    def author(self) -> hikari.User:
        """The user who invoked the command."""
        return self._message.author

    @property
    def clean_prefix(self) -> str:
        """
        The context's prefix, cleaned to remove user mentions. If the bot is stateless, then this just
        returns the raw prefix.
        """

        def replace(match):
            user = self.bot.cache.get_user(hikari.Snowflake(match.group(1)))
            return f"@{user}" if user is not None else self.prefix

        return re.sub(r"<@!?(\d+)>", replace, self.prefix)

    @property
    def guild(self) -> typing.Optional[hikari.Guild]:
        """
        The cached :obj:`hikari.Guild` instance for the context's guild ID.

        This will be None if the bot is stateless, the guild is not found in the cache,
        or the context is for a command run in DMs.
        """
        if self.guild_id is not None:
            return self.bot.cache.get_available_guild(self.guild_id)
        return None

    @property
    def channel(self) -> typing.Optional[hikari.TextableChannel]:
        """
        The cached :obj:`hikari.TextableChannel` instance for the context's channel ID.

        This will be None if the bot is stateless, the channel is not found in the cache,
        or the context is for a command run in DMs.
        """
        if self.guild_id is not None:
            return self.bot.cache.get_guild_channel(self.channel_id)
        return None

    @functools.wraps(hikari.Message.respond)
    async def respond(self, *args, **kwargs) -> hikari.Message:
        """
        Alias for ``ctx.message.respond(...)``.
        Replies to the message in the current context.

        Args:
            *args: The positional arguments :meth:`hikari.messages.Message.respond` is invoked with
            **kwargs: The keyword arguments :meth:`hikari.messages.Message.respond` is invoked with
        """
        return await self.message.respond(*args, **kwargs)

    async def send_help(self, obj: typing.Union[commands.Command, plugins.Plugin] = None) -> None:
        """
        Send help for the given object or the bot's help overview if no object
        is supplied to the current context.

        Args:
            obj (Union[ :obj:`~.commands.Command`, :obj:`~.plugins.Plugin` ]): The object to send help for.
                Defaults to ``None``.

        Returns:
            ``None``
        """
        await self.bot.send_help(self, obj)



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = [
    "CooldownStatus",
    "Bucket",
    "GlobalBucket",
    "UserBucket",
    "ChannelBucket",
    "GuildBucket",
    "CooldownManager",
    "cooldown",
    "dynamic_cooldown",
]

import abc
import time
import typing

from hikari.internal import enums

from lightbulb import errors
from lightbulb import utils

if typing.TYPE_CHECKING:
    from lightbulb import commands
    from lightbulb import context as context_


class CooldownStatus(int, enums.Enum):
    """The status of a cooldown bucket"""

    EXPIRED = 0
    """The cooldown bucket has expired"""

    INACTIVE = 1
    """The cooldown bucket timer has not been activated yet"""

    ACTIVE = 2
    """The cooldown bucket timer is currently active"""


class Bucket(abc.ABC):
    """
    Base class for representing a cooldown bucket. All buckets should inherit from this class.

    Args:
        length (:obj:`float`): Length of the cooldown timer.
        max_usages (:obj:`int`): Number of command usages before the cooldown is activated.
    """

    __slots__ = ("length", "usages", "commands_run", "activated", "start_time")

    def __init__(self, length: float, max_usages: int) -> None:
        self.length = length
        self.usages = max_usages
        self.commands_run: int = 0
        """Commands run for this bucket since it was created."""
        self.activated = False
        self.start_time: typing.Optional[float] = None
        """The start time of the bucket cooldown. This is relative to :meth:`time.perf_counter`."""

    @classmethod
    @abc.abstractmethod
    def extract_hash(cls, context: context_.Context) -> typing.Hashable:
        """
        Extracts the hash from the context which links a command usage to a single cooldown bucket.

        Args:
            context (:obj:`~.context.Context`): The context the command was invoked under.

        Returns:
            :obj:`typing.Hashable`: Hashable object linking the context to a cooldown bucket.
        """
        ...

    def acquire(self) -> CooldownStatus:
        """
        Get the current state of the cooldown and add a command usage if the cooldown is not
        currently active or has expired.

        Returns:
            :obj:`~CooldownStatus`: The status of the cooldown bucket.
        """
        if self.active:
            return CooldownStatus.ACTIVE
        elif self.activated and self.expired:
            return CooldownStatus.EXPIRED

        self.commands_run += 1
        if self.commands_run >= self.usages:
            self.activated = True
            self.start_time = time.perf_counter()
        return CooldownStatus.INACTIVE

    @property
    def active(self) -> bool:
        """
        Whether or not the cooldown represented by the bucket is currently active.
        """
        return self.activated and not self.expired

    @property
    def expired(self) -> bool:
        """
        Whether or not the cooldown represented by the bucket has expired.
        """
        if self.start_time is not None:
            return time.perf_counter() >= (self.start_time + self.length)
        return True


class GlobalBucket(Bucket):
    """
    Global cooldown bucket. All cooldowns will be applied globally if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: context_.Context) -> typing.Hashable:
        return 0


class UserBucket(Bucket):
    """
    User cooldown bucket. All cooldowns will be applied per user if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: context_.Context) -> typing.Hashable:
        return context.author.id


class ChannelBucket(Bucket):
    """
    Channel cooldown bucket. All cooldowns will be applied per channel if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: context_.Context) -> typing.Hashable:
        return context.channel_id


class GuildBucket(Bucket):
    """
    Guild cooldown bucket. All cooldowns will be applied per guild if you use this bucket. The command will
    still be permitted to be run in DMs in which case the DM channel ID will be used as the cooldown identifier
    instead of a guild ID.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: context_.Context) -> typing.Hashable:
        return context.guild_id if context.guild_id is not None else context.channel_id


class CooldownManager:
    """
    The cooldown manager for a command.

    Args:
        length (:obj:`float`): Length of the cooldown timer.
        usages (:obj:`int`): Number of command usages before the cooldown is activated.
        bucket (:obj:`Bucket`): The bucket that the cooldown should be evaluated under.
    """

    @typing.overload
    def __init__(self, length: float, usages: int, bucket: typing.Type[Bucket]) -> None:
        ...

    @typing.overload
    def __init__(self, *, callback: typing.Callable[[context_.Context], Bucket]) -> None:
        ...

    def __init__(
        self,
        length: typing.Optional[float] = None,
        usages: typing.Optional[int] = None,
        bucket: typing.Optional[typing.Type[Bucket]] = None,
        *,
        callback: typing.Optional[typing.Callable[[context_.Context], Bucket]] = None,
    ) -> None:
        if callback is not None:
            self.callback = callback
        elif length is not None and usages is not None and bucket is not None:
            self.length = length
            self.usages = usages
            self.bucket = bucket
        else:
            raise TypeError("Bad arguments...")
        self.cooldowns: typing.MutableMapping[typing.Hashable, Bucket] = {}
        """Mapping of a hashable to a :obj:`~Bucket` representing the currently stored cooldowns."""

    async def _get_bucket(self, context: context_.Context) -> Bucket:
        if not hasattr(self, "callback"):
            return self.bucket(self.length, self.usages)

        bucket = await utils.maybe_await(self.callback, context)

        if not isinstance(bucket, Bucket):
            raise TypeError("Bucket should derive the Bucket class")

        return bucket

    async def add_cooldown(self, context: context_.Context) -> None:
        """
        Add a cooldown under the given context. If an expired bucket already exists then it
        will be overwritten.

        Args:
            context (:obj:`~.context.Context`): The context to add a cooldown under.

        Returns:
            ``None``
        """
        bucket = await self._get_bucket(context)
        cooldown_hash = bucket.extract_hash(context)
        cooldown_bucket = self.cooldowns.get(cooldown_hash)
        if cooldown_bucket is not None:
            cooldown_status = cooldown_bucket.acquire()
            if cooldown_status == CooldownStatus.ACTIVE:
                # Cooldown has been activated
                raise errors.CommandIsOnCooldown(
                    "This command is on cooldown",
                    context.command,
                    (cooldown_bucket.start_time + cooldown_bucket.length) - time.perf_counter(),
                )
            elif cooldown_status == CooldownStatus.INACTIVE:
                # Cooldown has not yet been activated.
                return
        self.cooldowns[cooldown_hash] = bucket
        self.cooldowns[cooldown_hash].acquire()

    def reset_cooldown(self, context: context_.Context) -> None:
        """
        Reset the cooldown under the given context.

        Args:
            context (:obj:`~.context.Context`): The context to reset the cooldown under.

        Returns:
            ``None``
        """
        del self.cooldowns[self.bucket.extract_hash(context)]


def cooldown(
    length: float,
    usages: int,
    bucket: typing.Type[Bucket],
    *,
    manager_cls: typing.Type[CooldownManager] = CooldownManager,
):
    """
    Decorator which adds a cooldown to a command.
    Args:
        length (:obj:`float`): The amount of time before the cooldown expires.
        usages (:obj:`int`): The amount of usages of the command allowed before the cooldown activates.
        bucket (Type[ :obj:`~Bucket` ]): The bucket that the cooldown should be evaluated under.
    Keyword Args:
        manager_cls (Type[ :obj:`~CooldownManager` ]): The **uninstantiated** class to use as the command's
            cooldown manager. Defaults to :obj:`~CooldownManager`.
    Example:
        .. code-block:: python

            @lightbulb.cooldown(10, 1, lightbulb.UserBucket)
            @bot.command()
            async def ping(ctx):
                await ctx.respond("Pong!")

        This would make it so that each user can only use the ``ping`` command once every ten seconds.
    """

    def decorate(command: commands.Command) -> commands.Command:
        command.cooldown_manager = manager_cls(length, usages, bucket)
        return command

    return decorate


def dynamic_cooldown(
    callback: typing.Callable[[context_.Context], Bucket],
    *,
    manager_cls: typing.Type[CooldownManager] = CooldownManager,
):
    """
    Decorator which adds a more customized cooldown to a command.
    Args:
        callback (Callable[[ :obj:`~Context` ], :obj:`~Bucket`]): The callback that takes a Context object and returns a Bucket object.
    Keyword Args:
        manager_cls (Type[ :obj:`~CooldownManager` ]): The **uninstantiated** class to use as the command's
            cooldown manager. Defaults to :obj:`~CooldownManager`.
    Example:
        .. code-block:: python

            def callback(ctx):
                if ctx.author.id in ctx.bot.owner_ids:
                    return lightbulb.UserBucket(0, 1)

                return lightbulb.UserBucket(10, 1)

            @lightbulb.dynamic_cooldown(callback)
            @bot.command()
            async def ping(ctx):
                await ctx.respond("Pong!")

        This would make it so that owners bypass the cooldown and general users can only use the ``ping`` command once every ten seconds.
    """

    def decorate(command: commands.Command) -> commands.Command:
        command.cooldown_manager = manager_cls(callback=callback)
        return command

    return decorate



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = [
    "SlashCommandBase",
    "TopLevelSlashCommandBase",
    "SlashCommand",
    "SlashCommandGroup",
    "SlashSubGroup",
    "SlashSubCommand",
]

import abc
import logging
import typing

import hikari

if typing.TYPE_CHECKING:
    from lightbulb import command_handler
    from lightbulb.slash_commands import context as context_

_LOGGER = logging.getLogger("lightbulb")


def _resolve_subcommand(
    context: context_.SlashCommandContext, subcommands: typing.MutableMapping[str, SlashCommandBase]
) -> typing.Optional[str]:
    option_name = None
    for option in context.options:
        if option in subcommands:
            option_name = option
            break
    return option_name


class SlashCommandBase(abc.ABC):
    """
    Abstract base class for slash command-like classes.

    Args:
        bot (:obj:`~lightbulb.command_handler.Bot`): The bot instance the command will be added to.
    """

    __slots__: typing.Sequence[str] = ("bot", "_instances")

    def __init__(self, bot: command_handler.Bot) -> None:
        self.bot = bot
        self._instances: typing.MutableMapping[hikari.Snowflakeish, hikari.Command] = {}

    @abc.abstractmethod
    async def __call__(self, context: context_.SlashCommandContext) -> None:
        ...

    @property
    def name(self) -> str:
        """
        The name of the slash command. Defaults to the class name converted to lowercase. Can be a maximum
        of 32 characters long.
        """
        return self.__class__.__name__.lower()

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """
        The description of the slash command. Can be a maximum of 100 characters long.
        """
        ...


class TopLevelSlashCommandBase(SlashCommandBase, abc.ABC):
    """
    Abstract base class for top level slash commands and slash command groups.
    """

    @property
    @abc.abstractmethod
    def enabled_guilds(self) -> typing.Optional[hikari.SnowflakeishSequence]:
        """
        The guilds that the slash command is enabled in. If ``None``, the command will be added as
        a global command.
        """
        ...

    @abc.abstractmethod
    async def create(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> hikari.Command:
        """
        Creates the command for a specific guild, or globally if no guild ID is given.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to add the command to.
            guild_id (Optional[:obj:`int`]): The ID of the guild to create the command for.

        Returns:
            :obj:`hikari.Command`: The command object created.
        """
        ...

    async def delete(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> None:
        """
        Deletes the command for a specific guild, or globally if no guild ID is given.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to delete the command from.
            guild_id (Optional[hikari.Snowflakeish]): The ID of the guild to delete the command for.

        Returns:
            ``None``
        """
        await self.bot.rest.delete_application_command(
            app, self._instances[guild_id], **({"guild": guild_id} if guild_id is not None else {})
        )
        self._instances.pop(guild_id)

    async def auto_create(self, app: hikari.SnowflakeishOr[hikari.PartialApplication]) -> None:
        """
        Creates the command in the appropriate scopes (guilds, global) automatically.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to add the command to.

        Returns:
            ``None``
        """
        if self.enabled_guilds is None:
            await self.create(app)
        else:
            for guild_id in self.enabled_guilds:
                await self.create(app, guild_id)

    async def auto_delete(self, app: hikari.SnowflakeishOr[hikari.PartialApplication]) -> None:
        """
        Deletes the command from the appropriate scopes (guilds, global) automatically.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to remove the command from.

        Returns:
            ``None``
        """
        if self.enabled_guilds is None:
            await self.delete(app)
        else:
            for guild_id in self.enabled_guilds:
                await self.delete(app, guild_id)

    def get_command(self, guild_id: typing.Optional[hikari.Snowflakeish] = None) -> typing.Optional[hikari.Command]:
        """
        Gets the :obj:`hikari.Command` instance of this command class for a given ``guild_id``, or the global
        instance if no ``guild_id`` is provided. Returns ``None`` if no instance is found for the given ``guild_id``.

        Args:
            guild_id (Optional[:obj:`hikari.Snowflakeish`): The guild ID to get the :obj:`hikari.Command` instance
                for.

        Returns:
            Optional[:obj:`hikari.Command`]: Command instance for that guild, or ``None`` if no instance exists.
        """
        return self._instances.get(guild_id)


class SlashCommand(TopLevelSlashCommandBase, abc.ABC):
    """
    Abstract base class for top level slash commands. All slash commands that are not groups
    should inherit from this class.

    All abstract methods **must** be implemented by your custom slash command class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.SlashCommandBase.description`
    - :obj:`~lightbulb.slash_commands.SlashCommand.options`
    - :obj:`~lightbulb.slash_commands.TopLevelSlashCommandBase.enabled_guilds`
    - :obj:`~lightbulb.slash_commands.SlashCommand.callback`
    """

    __slots__: typing.Sequence[str] = ()

    async def __call__(self, context: context_.SlashCommandContext) -> None:
        return await self.callback(context)

    async def create(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> hikari.Command:
        created_command = await self.bot.rest.create_application_command(
            app,
            self.name,
            self.description,
            options=self.options,
            **({"guild": guild_id} if guild_id is not None else {}),
        )
        self._instances[guild_id] = created_command
        return created_command

    @property
    @abc.abstractmethod
    def options(self) -> typing.Sequence[hikari.CommandOption]:
        """
        The slash command options. Can include up to a maximum of 25 options.
        """
        ...

    @abc.abstractmethod
    async def callback(self, context: context_.SlashCommandContext) -> None:
        """
        The slash command callback method. This method will be called whenever the slash command is invoked.

        Args:
            context (:obj:`~lightbulb.slash_commands.SlashCommandContext`): The context that the slash command
                was invoked under.

        Returns:
            ``None``
        """
        ...


class SlashSubCommand(SlashCommandBase, abc.ABC):
    """
    Abstract base class for slash subcommands. All slash subcommands should inherit from this class.

    All abstract methods **must** be implemented by your custom slash subcommand class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.SlashCOmmandBase.description`
    - :obj:`~lightbulb.slash_commands.SlashSubCommand.options`
    - :obj:`~lightbulb.slash_commands.SlashSubCommand.callback`
    """

    async def __call__(self, context: context_.SlashCommandContext) -> None:
        return await self.callback(context)

    def as_option(self) -> hikari.CommandOption:
        """
        Creates and returns the appropriate :obj:`hikari.CommandOption` representation for this
        subcommand class.

        Returns:
            :obj:`hikari.CommandOption`: The ``CommandOption`` version of the subcommand class.
        """
        return hikari.CommandOption(
            name=self.name,
            description=self.description,
            type=hikari.OptionType.SUB_COMMAND,
            options=self.options,
            is_required=False,
        )

    @abc.abstractmethod
    async def callback(self, context: context_.SlashCommandContext) -> None:
        """
        The subcommand callback method. This method will be called whenever the  subcommand is invoked.

        Args:
            context (:obj:`~lightbulb.slash_commands.SlashCommandContext`): The context that the subcommand
                was invoked under. The :obj:`~lightbulb.slash_commands.SlashCommandContext.options` attribute
                will have been replaced by the options that the subcommand was invoked with, instead of those that
                the command as a whole was invoked with  (they can still be accessed through
                :obj:`~lightbulb.slash_commands.SlashCommandContext.interaction.options` if necessary).

        Returns:
            ``None``
        """
        ...

    @property
    @abc.abstractmethod
    def options(self) -> typing.Sequence[hikari.CommandOption]:
        """
        The subcommand options. Can include up to a maximum of 25 options.
        """
        ...


class SlashSubGroup(SlashCommandBase, abc.ABC):
    """
    Abstract base class for slash subgroups. All slash subgroups should inherit from this class.

    All abstract methods **must** be implemented by your custom slash subgroup class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.SlashCommandBase.description`
    """

    __slots__: typing.Sequence[str] = ("_subcommands",)

    _subcommand_list: typing.List[typing.Type[SlashSubCommand]] = []

    def __init__(self, bot: command_handler.Bot):
        super().__init__(bot)
        self._subcommands: typing.MutableMapping[str, SlashSubCommand] = {}
        for cmd_class in self._subcommand_list:
            cmd = cmd_class(bot)
            self._subcommands[cmd.name] = cmd

    async def __call__(self, context: context_.SlashCommandContext) -> None:
        option_name = _resolve_subcommand(context, self._subcommands)
        if option_name is None:
            return

        _LOGGER.debug("invoking slash subcommand %s", option_name)
        # Replace the context options with the options for the subcommand, the old options
        # can still be accessed through context.interaction.options
        new_options = context.options[option_name].options
        context.options = {option.name: option for option in new_options} if new_options is not None else {}
        return await self._subcommands[option_name](context)

    @classmethod
    def subcommand(cls) -> typing.Callable[[typing.Type[SlashSubCommand]], typing.Type[SlashSubCommand]]:
        """
        Decorator which registers a subcommand to this slash command group.
        """

        def decorate(subcommand_class: typing.Type[SlashSubCommand]) -> typing.Type[SlashSubCommand]:
            cls._subcommand_list.append(subcommand_class)
            return subcommand_class

        return decorate

    def as_option(self):
        """
        Creates and returns the appropriate :obj:`hikari.CommandOption` representation for this
        subgroup class.

        Returns:
            :obj:`hikari.CommandOption`: The ``CommandOption`` version of the subgroup class.
        """
        return hikari.CommandOption(
            name=self.name,
            description=self.description,
            type=hikari.OptionType.SUB_COMMAND_GROUP,
            options=[cmd.as_option() for cmd in self._subcommands.values()],
            is_required=False,
        )


class SlashCommandGroup(TopLevelSlashCommandBase, abc.ABC):
    """
    Abstract base class for slash command groups. All slash command groups should inherit from this class.

    All abstract methods **must** be implemented by your custom slash command group class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.SlashCommandBase.description`
    - :obj:`~lightbulb.slash_commands.TopLevelSlashCommandBase.enabled_guilds`
    """

    __slots__: typing.Sequence[str] = ("_subcommands",)

    _subcommand_list: typing.List[typing.Union[typing.Type[SlashSubCommand], typing.Type[SlashSubGroup]]] = []

    def __init__(self, bot: command_handler.Bot):
        super().__init__(bot)
        self._subcommands: typing.MutableMapping[str, SlashSubCommand] = {}
        for cmd_class in self._subcommand_list:
            cmd = cmd_class(bot)
            self._subcommands[cmd.name] = cmd

    async def __call__(self, context: context_.SlashCommandContext) -> None:
        option_name = _resolve_subcommand(context, self._subcommands)
        if option_name is None:
            return

        _LOGGER.debug("invoking slash subcommand %s", option_name)
        # Replace the context options with the options for the subcommand, the old options
        # can still be accessed through context.interaction.options
        new_options = context.options[option_name].options
        context.options = {option.name: option for option in new_options} if new_options is not None else {}
        return await self._subcommands[option_name](context)

    @classmethod
    def subcommand(cls) -> typing.Callable[[typing.Type[SlashSubCommand]], typing.Type[SlashSubCommand]]:
        """
        Decorator which registers a subcommand to this slash command group.
        """

        def decorate(subcommand_class: typing.Type[SlashSubCommand]) -> typing.Type[SlashSubCommand]:
            cls._subcommand_list.append(subcommand_class)
            return subcommand_class

        return decorate

    @classmethod
    def subgroup(cls) -> typing.Callable[[typing.Type[SlashSubGroup]], typing.Type[SlashSubGroup]]:
        """
        Decorator which registers a subgroup to this slash command group.
        """

        def decorate(subgroup_class: typing.Type[SlashSubGroup]) -> typing.Type[SlashSubGroup]:
            cls._subcommand_list.append(subgroup_class)
            return subgroup_class

        return decorate

    async def create(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> hikari.Command:
        created_command = await self.bot.rest.create_application_command(
            app,
            self.name,
            self.description,
            options=[cmd.as_option() for cmd in self._subcommands.values()],
            **({"guild": guild_id} if guild_id is not None else {}),
        )
        self._instances[guild_id] = created_command
        return created_command



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = ["StringView"]

import typing

from lightbulb import errors

_quotes = {
    '"': '"',
    "'": "'",
    "‘": "’",
    "‚": "‛",
    "“": "”",
    "„": "‟",
    "⹂": "⹂",
    "「": "」",
    "『": "』",
    "〝": "〞",
    "﹁": "﹂",
    "﹃": "﹄",
    "＂": "＂",
    "｢": "｣",
    "«": "»",
    "‹": "›",
    "《": "》",
    "〈": "〉",
}


class StringView:
    """
    Class to decompose a string (generally message content) into its arguments.

    Args:
        raw_string (:obj:`str`): The string to break down into its arguments.
    """

    __slots__: typing.Sequence[str] = ("text", "index", "final_args", "current_arg", "expect_quote")

    def __init__(self, raw_string: str) -> None:
        self.text = raw_string
        self.index = 0
        self.final_args = []
        self.current_arg = []
        self.expect_quote = None

    def _next_str(self) -> str:
        buff = []
        previous = None
        while self.index < len(self.text):
            char = self.text[self.index]
            try:
                if char.isspace() and self.expect_quote is None:
                    self.index += 1
                    if not buff:
                        continue
                    return "".join(buff)
                elif not self.expect_quote and char in _quotes:
                    if previous and not previous.isspace():
                        raise errors.UnexpectedQuotes(char)
                    self.expect_quote = _quotes[char]
                    self.index += 1
                    continue
                elif char == self.expect_quote:
                    self.index += 1
                    if self.index < len(self.text) and not (next_char := self.text[self.index]).isspace():
                        raise errors.ExpectedSpaces(next_char)
                    self.expect_quote = None
                    return "".join(buff)
                elif char == "\\":
                    self.index += 1
                    if self.index < len(self.text):
                        buff.append(self.text[self.index])
                        self.index += 1
                        continue
                    raise errors.PrematureEOF()
                else:
                    buff.append(char)
                    self.index += 1
            finally:
                previous = char

        if self.expect_quote:
            raise errors.UnclosedQuotes("".join(buff))
        if buff:
            return "".join(buff)

    def deconstruct_str(self, *, max_parse: typing.Optional[int] = None) -> typing.Tuple[typing.List[str], str]:
        """
        Method called to deconstruct the string passed in at instantiation into
        a list of all of its arguments. Arguments are defined as words separated by whitespace,
        or as words enclosed in quotation marks.

        Keyword Args:
            max_parse (Optional[ :obj:`int` ]): The maximum number of arguments to parse. If unspecified,
                as many arguments will be parsed as possible.

        Raises:
            :obj:`~.errors.UnclosedQuotes`: When an unexpected EOF is encountered.

        Returns:
            Tuple[ List[ :obj:`str` ], :obj:`str`]: The arguments extracted from the string, as well as any
                remainder if ``max_parse`` was supplied.
        """
        max_parse = max_parse if max_parse is not None else float("inf")

        finished = False
        args_list = []
        while not finished and len(args_list) < max_parse:
            arg = self._next_str()
            if arg is not None:
                args_list.append(arg)
            else:
                finished = True
        return args_list, self.text[self.index :].strip()



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.Tuple[str]] = [
    "get_help_text",
    "get_command_signature",
    "filter_commands",
    "HelpCommand",
]

import inspect
import typing

from lightbulb import commands
from lightbulb import errors
from lightbulb import plugins
from lightbulb import utils
from lightbulb.converters import _ConsumeRestConverter
from lightbulb.converters import _DefaultingConverter

if typing.TYPE_CHECKING:
    from lightbulb import command_handler
    from lightbulb import context as context_


def _format_help_text(help_text: str) -> str:
    segments = help_text.split("\n\n")
    return "\n".join(seg.replace("\n", " ").strip() for seg in segments)


def get_help_text(obj: typing.Union[commands.Command, plugins.Plugin]) -> str:
    """
    Get the help text for a command, group or plugin, extracted from its docstring.

    Args:
        obj (Union[ :obj:`~.commands.Command`, :obj:`~.commands.Group`, :obj:`~.plugins.Plugin` ]): The
            object to get the help text for.

    Returns:
        :obj:`str`: The extracted help text, or an empty string if no help text has
        been provided for the object.
    """
    if not isinstance(obj, plugins.Plugin):
        doc = inspect.getdoc(obj._callback)
        return _format_help_text(doc if doc is not None else "")
    else:
        doc = inspect.getdoc(obj)
        return _format_help_text(doc if doc != inspect.getdoc(plugins.Plugin) else "")


def get_command_signature(command: commands.Command) -> str:
    """
    Get the command signature (usage) for a command or command group.
    The signature is returned in the format:
    ``<command name> <required arg> [optional arg]``

    Args:
        command (:obj:`~.commands.Command`): The command or group to get the signature for.

    Returns:
        :obj:`str`: Signature for the command.
    """
    items = [command.qualified_name]
    for argname, converter in zip(command.arg_details.arguments, command.arg_details.converters):
        if isinstance(converter, _DefaultingConverter):
            default = converter.default
            if isinstance(default, dict):
                default = default.get(argname)

            items.append(f"[{argname}={default}]")
        else:
            items.append(f"<{argname}>")

        if isinstance(converter, _ConsumeRestConverter):
            break
    return " ".join(items)


async def filter_commands(
    context: context_.Context,
    command_list: typing.Iterable[commands.Command],
) -> typing.List[commands.Command]:
    """
    Filter a list of :obj:`~.commands.Command` and :obj:`~.commands.Group`, removing any commands that cannot
    be run under the given context by running all checks for each command in turn.

    Args:
        context (:obj:`~.context.Context`): The context to filter the commands under.
        command_list (List[ :obj:`~.commands.Command` ]): List of commands to filter.

    Returns:
        List[ :obj:`~.commands.Command` ]: List containing filtered commands.
    """
    filtered_commands = set()
    for command in command_list:
        if command.hidden:
            continue
        try:
            await command.is_runnable(context)
            filtered_commands.add(command)
        except errors.CheckFailure:
            pass
    return list(filtered_commands)


@commands.command(name="help")
async def _help_cmd(ctx: context_.Context) -> None:
    """
    Displays help for the bot, a command, or a category.

    If no object is specified with the command then a help menu
    for the bot as a whole is displayed instead.
    """
    obj = ctx.message.content[len(f"{ctx.prefix}{ctx.invoked_with}") :].strip().split()
    await ctx.bot.help_command.resolve_help_obj(ctx, obj)


class HelpCommand:
    """
    The default help command implementation. This class should be subclassed if you
    wish to customise the format or other aspects of the bot's help command.

    Args:
        bot (:obj:`~.command_handler.Bot`): Bot instance to add the help command class to.
    """

    def __init__(self, bot: command_handler.Bot) -> None:
        self.bot = bot
        if self.bot.get_command("help") is None:
            self.bot.add_command(_help_cmd)

    @staticmethod
    async def send_paginated_help(text: typing.Sequence[str], context: context_.Context) -> None:
        """
        Paginate the help text using :obj:`~.utils.pag.StringPaginator` and send each
        of the created pages to the context in order. Note that by default, only the
        help overview is paginated as this is the most likely to exceed the 2000
        character message limit.

        Args:
            text (Sequence[ :obj:`str` ]): A sequence of text to be paginated.
            context (:obj:`~.context.Context`): The context to send the help to.

        Returns:
            ``None``
        """
        pag = utils.pag.StringPaginator()
        for line in text:
            pag.add_line(line)
        for page in pag.build_pages():
            await context.respond(page)

    async def resolve_help_obj(self, context: context_.Context, obj: typing.List[str]) -> None:
        """
        Resolve the object to send help information for from the
        arguments passed to the help command.

        Args:
            context (:obj:`~.context.Context`): Context to send the help to.
            obj (List[ :obj:`str` ]): Arguments supplied to the help command.

        Returns:
            ``None``
        """
        if not obj:
            await self.send_help_overview(context)
        elif len(obj) == 1:
            command = self.bot.get_command(obj[0])
            if isinstance(command, commands.Group):
                await self.send_group_help(context, command)
            elif isinstance(command, commands.Command):
                await self.send_command_help(context, command)
            else:
                plugin = self.bot.get_plugin(obj[0])
                if plugin is not None:
                    await self.send_plugin_help(context, plugin)
                else:
                    await self.object_not_found(context, obj[0])
        else:
            if (plugin := self.bot.get_plugin(" ".join(obj))) is not None:
                await self.send_plugin_help(context, plugin)
                return

            command = self.bot.get_command(obj[0])

            next_obj = 1
            while isinstance(command, commands.Group):
                try:
                    command = command.get_subcommand(obj[next_obj])
                    next_obj += 1
                except IndexError:
                    break

            if isinstance(command, commands.Group):
                await self.send_group_help(context, command)
            elif isinstance(command, commands.Command):
                await self.send_command_help(context, command)
            else:
                await self.object_not_found(context, " ".join(obj))

    async def object_not_found(self, context: context_.Context, name: str) -> None:
        """
        Method called when help is requested for an object that does not exist.

        Args:
            context (:obj:`~.context.Context`): Context to send the error message to.
            name (:obj:`str`): The name of the object that help was requested for.

        Returns:
            ``None``
        """
        await context.respond(f"`{name}` is not a valid command, group or category.")

    async def send_help_overview(self, context: context_.Context) -> None:
        """
        Method called when the help command is run without any arguments.

        Args:
            context (:obj:`~.context.Context`): Context to send the help to.

        Returns:
            ``None``
        """
        plugin_commands = [
            [plugin.name, await filter_commands(context, plugin._commands.values())]
            for plugin in self.bot.plugins.values()
        ]
        all_plugin_commands = []
        for _, cmds in plugin_commands:
            all_plugin_commands.extend(cmds)
        uncategorised_commands = await filter_commands(context, self.bot.commands.difference(set(all_plugin_commands)))
        plugin_commands.insert(0, ["Uncategorised", uncategorised_commands])

        help_text = ["> __**Bot help**__\n"]
        for plugin, commands in plugin_commands:
            if not commands:
                continue
            help_text.append(f"> **{plugin}**")
            for c in sorted(commands, key=lambda c: c.name):
                short_help = get_help_text(c).split("\n")[0]
                help_text.append(f"> • `{c.name}` - {short_help}")
        help_text.append(f"> \n> Use `{context.clean_prefix}help [command]` for more information.")
        await self.send_paginated_help(help_text, context)

    async def send_plugin_help(self, context: context_.Context, plugin: plugins.Plugin) -> None:
        """
        Method called when the help command is run with an argument that
        resolves into the name of a plugin.

        Args:
            context (:obj:`~.context.Context`): Context to send the help to.
            plugin (:obj:`~.plugins.Plugin`): Plugin object to send help for.

        Returns:
            ``None``
        """
        help_text = [
            f"> **Help for category `{plugin.name}`**",
            get_help_text(plugin).replace("\n", "\n> ") or "No help text provided.",
            f"Commands:",
            ", ".join(f"`{c.name}`" for c in sorted(plugin._commands.values(), key=lambda c: c.name))
            or "No commands in the category",
        ]
        await context.respond("\n> ".join(help_text))

    async def send_command_help(self, context: context_.Context, command: commands.Command) -> None:
        """
        Method called when the help command is run with an argument that
        resolves into the name of a registered command.

        Args:
            context (:obj:`~.context.Context`): Context to send the help to.
            command (:obj:`~.commands.Command`): Command object to send help for.

        Returns:
            ``None``
        """
        help_text = [
            f"> **Help for command `{command.name}`**",
            f"Usage:",
            f"```{context.clean_prefix}{get_command_signature(command)}```",
            get_help_text(command).replace("\n", "\n> ") or "No help text provided.",
        ]
        await context.respond("\n> ".join(help_text))

    async def send_group_help(self, context: context_.Context, group: commands.Group) -> None:
        """
        Method called when the help command is run with an argument that
        resolves into the name of a registered command group.

        Args:
            context (:obj:`~.context.Context`): Context to send the help to.
            command (:obj:`~.commands.Group`): Group object to send help for.

        Returns:
            ``None``
        """
        help_text = [
            f"> **Help for command group `{group.name}`**",
            "Usage:",
            f"```{context.clean_prefix}{get_command_signature(group)}```",
            get_help_text(group).replace("\n", "\n> ") or "No help text provided.",
            f"Subcommands: {', '.join(f'`{c.name}`' for c in sorted(group.subcommands, key=lambda c: c.name))}"
            if group.subcommands
            else "No subcommands in the group",
        ]
        await context.respond("\n> ".join(help_text))



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.

import inspect
import typing

from lightbulb.utils import nav
from lightbulb.utils import pag
from lightbulb.utils import search
from lightbulb.utils.nav import *
from lightbulb.utils.pag import *
from lightbulb.utils.search import *

__all__: typing.Final[typing.List[str]] = [*nav.__all__, *pag.__all__, *search.__all__, "maybe_await"]


async def maybe_await(callable_, *args, **kwargs):
    result = callable_(*args, **kwargs)
    if inspect.iscoroutine(result):
        result = await result
    return result



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = [
    "LightbulbError",
    "ExtensionError",
    "ExtensionAlreadyLoaded",
    "ExtensionNotLoaded",
    "ExtensionMissingLoad",
    "ExtensionMissingUnload",
    "CommandError",
    "CommandNotFound",
    "NotEnoughArguments",
    "TooManyArguments",
    "ConverterFailure",
    "CommandIsOnCooldown",
    "CommandSyntaxError",
    "PrematureEOF",
    "UnclosedQuotes",
    "CheckFailure",
    "OnlyInGuild",
    "OnlyInDM",
    "NotOwner",
    "BotOnly",
    "HumanOnly",
    "NSFWChannelOnly",
    "MissingRequiredRole",
    "MissingRequiredPermission",
    "BotMissingRequiredPermission",
    "MissingRequiredAttachment",
    "CommandInvocationError",
]

import abc
import typing

import hikari

from lightbulb import commands


class LightbulbError(Exception):
    """Base for any exception raised by lightbulb."""


class ExtensionError(LightbulbError):
    """Base exception for errors incurred during the loading and unloading of extensions."""

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class ExtensionAlreadyLoaded(ExtensionError):
    """Exception raised when an extension already loaded is attempted to be loaded."""


class ExtensionNotLoaded(ExtensionError):
    """Exception raised when an extension not already loaded is attempted to be unloaded."""


class ExtensionMissingLoad(ExtensionError):
    """Exception raised when an extension is attempted to be loaded but does not contain a load function"""


class ExtensionMissingUnload(ExtensionError):
    """Exception raised when an extension is attempted to be unloaded but does not contain an unload function"""


class CommandError(LightbulbError):
    """Base exception for errors incurred during handling of commands."""


class CommandNotFound(CommandError):
    """
    Exception raised when a command when attempted to be invoked but one with that name could not be found.
    """

    def __init__(self, invoked_with: str) -> None:
        self.invoked_with: str = invoked_with
        """The command string that was attempted to be invoked."""


class NotEnoughArguments(CommandError):
    """
    Exception raised when a command is run without a sufficient number of arguments.
    """

    def __init__(self, command: commands.Command, missing_args: typing.List[str]) -> None:
        self.command: commands.Command = command
        """The command string that was attempted to be invoked."""
        self.missing_args: typing.List[str] = missing_args
        """The required arguments that are missing."""


class TooManyArguments(CommandError):
    """
    Exception raised when a command is run with too many arguments, and the command has been
    defined to not accept any extra arguments when invoked.
    """

    def __init__(self, command: commands.Command) -> None:
        self.command: commands.Command = command
        """The command string that was attempted to be invoked."""


class ConverterFailure(CommandError):
    """
    Exception raised when a converter for a command argument fails.
    """

    def __init__(self, text: typing.Optional[str] = None) -> None:
        self.text: typing.Optional[str] = text
        """The error text."""


class CommandIsOnCooldown(CommandError):
    """
    Exception raised when a command is attempted to be run but is currently on cooldown.
    """

    def __init__(self, text: str, command: commands.Command, retry_in: float) -> None:
        self.text: str = text
        """The error text."""

        self.command: commands.Command = command
        """The command that is on cooldown."""

        self.retry_in: float = retry_in
        """Number of seconds remaining for the cooldown."""


class CommandSyntaxError(CommandError, abc.ABC):
    """
    Base error raised if a syntax issue occurs parsing invocation arguments.
    """

    # Forces the class to be abstract.
    @abc.abstractmethod
    def __init__(self) -> None:
        ...


class PrematureEOF(CommandSyntaxError):
    """
    Error raised if EOF (end of input) was reached, but more content was
    expected.
    """

    def __init__(self) -> None:
        # Required to override the abstract super init.
        super().__init__()


class UnclosedQuotes(CommandSyntaxError):
    """
    Error raised when no closing quote is found for a quoted argument.
    """

    def __init__(self, text: str) -> None:
        # Required to override the abstract super init.
        super().__init__()
        self.text = text
        """The text that caused the error to be raised."""


class UnexpectedQuotes(CommandSyntaxError):
    """
    Error raised when a quote mark is found in non-quoted string.
    """

    def __init__(self, quote: str) -> None:
        super().__init__()
        self.quote = quote
        """The quote mark that caused the error to be raised."""


class ExpectedSpaces(CommandSyntaxError):
    """
    Error raised when no spaces found in the end of a quoted string
    """

    def __init__(self, char: str) -> None:
        super().__init__()
        self.char = char
        """The character that's expected to be a space character"""


class CheckFailure(CommandError):
    """
    Base error that is raised when a check fails for a command. Anything raised by a check
    should inherit from this class.
    """

    def __init__(self, text: typing.Optional[str] = None) -> None:
        self.text: typing.Optional[str] = text
        """The error text."""


class OnlyInGuild(CheckFailure):
    """
    Error raised when a command marked as guild only is attempted to be invoked in DMs.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class OnlyInDM(CheckFailure):
    """
    Error raised when a command marked as DM only is attempted to be invoked in a guild.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class NotOwner(CheckFailure):
    """
    Error raised when a command marked as owner only is attempted to be invoked by another user.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class BotOnly(CheckFailure):
    """
    Error raised when the command invoker is not a bot.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class WebhookOnly(CheckFailure):
    """
    Error raised when the command invoker is not a webhook.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class HumanOnly(CheckFailure):
    """
    Error raised when the command invoker is not an human.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class NSFWChannelOnly(CheckFailure):
    """
    Error raised when a command that must be invoked in an NSFW channel is attempted to be invoked outside of one.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class MissingRequiredRole(CheckFailure):
    """
    Error raised when the member invoking a command is missing one or more role required.
    """

    def __init__(self, text: str) -> None:
        self.text: str = text
        """The error text."""


class MissingRequiredPermission(CheckFailure):
    """
    Error raised when the member invoking a command is missing one or more permission required.
    """

    def __init__(self, text: str, permissions: hikari.Permissions) -> None:
        self.text: str = text
        """The error text."""
        self.permissions: hikari.Permissions = permissions
        """Permission(s) the bot is missing."""


class BotMissingRequiredPermission(CheckFailure):
    """
    Error raised when the bot is missing one or more permission required for the command to be run.
    """

    def __init__(self, text: str, permissions: hikari.Permissions) -> None:
        self.text: str = text
        """The error text."""
        self.permissions: hikari.Permissions = permissions
        """Permission(s) the bot is missing."""


class MissingRequiredAttachment(CheckFailure):
    """
    Error raised when the command invocation message is missing an attachment, or an
    attachment with the correct file extension.
    """

    def __init__(self, text) -> None:
        self.text: str = text
        """The error text."""


class CommandInvocationError(CommandError):
    """
    Error raised if an error is encountered during command invocation. This will only be raised
    if all the checks passed and an error was raised somewhere inside the command.
    This effectively acts as a wrapper for the original exception for easier handling in an error handler.
    """

    def __init__(self, text: str, original: Exception) -> None:
        self.text: str = text
        """The error text."""
        self.original: Exception = original
        """The original exception that caused this to be raised."""

    @property
    def __cause__(self) -> Exception:
        return self.original



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.List[str] = [
    "StringNavigator",
    "EmbedNavigator",
    "Navigator",
    "NavButton",
    "next_page",
    "prev_page",
    "first_page",
    "last_page",
    "stop",
]

import abc
import asyncio
import typing

import hikari

from lightbulb.context import Context


class NavButton:
    """
    A navigator button. Contains the emoji linked to the button as well as
    the coroutine to be called when the button is pressed.

    Args:
        emoji (Union[ :obj:`str`, :obj:`~hikari.emojis.Emoji` ]): The emoji linked to the button.
        callback: The coroutine function to be called on button press.
    """

    __slots__ = ("emoji", "callback")

    def __init__(
        self,
        emoji: typing.Union[str, hikari.Emoji],
        callback: typing.Callable[[Navigator, hikari.ReactionAddEvent], typing.Coroutine[typing.Any, typing.Any, None]],
    ) -> None:
        if isinstance(emoji, str):
            emoji = hikari.Emoji.parse(emoji)
        self.emoji = emoji
        self.callback = callback

    def is_pressed(self, event: hikari.ReactionAddEvent) -> bool:
        """
        Check if the button is pressed in a given event.

        Args:
            event (:obj:`~hikari.events.message.MessageReactionEvent`): The event to check the button is pressed in.

        Returns:
            :obj:`bool`: Whether or not the button is pressed in the given event.
        """
        if event.emoji_id is not None and isinstance(self.emoji, hikari.CustomEmoji):
            return event.emoji_id == self.emoji.id
        return event.emoji_name == self.emoji.name

    def press(self, nav: Navigator, event: hikari.ReactionAddEvent) -> typing.Coroutine[typing.Any, typing.Any, None]:
        """
        Call the button's callback coroutine and return the awaitable.

        Returns:
            Coroutine[``None``, Any, ``None``]: Returned awaitable from the coroutine call.
        """
        return self.callback(nav, event)


T = typing.TypeVar("T")


async def next_page(nav: Navigator, _) -> None:
    """
    :obj:`NavButton` callback to make the navigator go to the next page.
    """
    nav.current_page_index += 1
    nav.current_page_index %= len(nav.pages)


async def prev_page(nav: Navigator, _) -> None:
    """
    :obj:`NavButton` callback to make the navigator go to the previous page.
    """
    nav.current_page_index -= 1
    if nav.current_page_index < 0:
        nav.current_page_index = len(nav.pages) - 1


async def first_page(nav: Navigator, _) -> None:
    """
    :obj:`NavButton` callback to make the navigator go to the first page.
    """
    nav.current_page_index = 0


async def last_page(nav: Navigator, _) -> None:
    """
    :obj:`NavButton` callback to make the navigator go to the last page.
    """
    nav.current_page_index = len(nav.pages) - 1


async def stop(nav: Navigator, _) -> None:
    """
    :obj:`NavButton` callback to make the navigator stop navigation.
    """
    nav._msg.app.unsubscribe(hikari.ReactionAddEvent, nav._process_reaction_add)
    await nav._msg.delete()
    nav._msg = None


class Navigator(abc.ABC, typing.Generic[T]):
    def __init__(
        self,
        pages: typing.Union[typing.Iterable[T], typing.Iterator[T]],
        *,
        buttons: typing.Optional[typing.Sequence[NavButton]] = None,
        timeout: float = 120,
    ) -> None:
        if not pages:
            raise ValueError("You cannot pass fewer than 1 page to the navigator.")
        self.pages: typing.Sequence[T] = tuple(pages)

        if buttons is not None:
            if any(not isinstance(btn, NavButton) for btn in buttons):
                raise TypeError("Buttons must be an instance of NavButton")

        if len(self.pages) == 1 and not buttons:
            self.buttons = [NavButton("\N{BLACK SQUARE FOR STOP}\N{VARIATION SELECTOR-16}", stop)]
        else:
            self.buttons: typing.Sequence[NavButton] = buttons if buttons is not None else self.create_default_buttons()

        self._timeout: float = timeout
        self.current_page_index: int = 0
        self._context: typing.Optional[Context] = None
        self._msg: typing.Optional[hikari.Message] = None
        self._timeout_task = None

    @abc.abstractmethod
    async def _edit_msg(self, message: hikari.Message, page: T) -> hikari.Message:
        ...

    @abc.abstractmethod
    async def _send_initial_msg(self, page: T) -> hikari.Message:
        ...

    def create_default_buttons(self) -> typing.Sequence[NavButton]:
        buttons = (
            NavButton("\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\N{VARIATION SELECTOR-16}", first_page),
            NavButton("\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}", prev_page),
            NavButton("\N{BLACK SQUARE FOR STOP}\N{VARIATION SELECTOR-16}", stop),
            NavButton("\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}", next_page),
            NavButton("\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\N{VARIATION SELECTOR-16}", last_page),
        )
        return buttons

    async def _process_reaction_add(self, event: hikari.ReactionAddEvent) -> None:
        if (
            event.user_id != self._context.message.author.id
            or event.channel_id != self._context.channel_id
            or event.message_id != self._msg.id
        ):
            return

        for button in self.buttons:
            if button.is_pressed(event):
                await button.press(self, event)
                if self._msg is not None:
                    await self._edit_msg(self._msg, self.pages[self.current_page_index])
                    try:
                        await self._msg.remove_reaction(button.emoji, user=self._context.author)
                    except hikari.ForbiddenError:
                        pass
                break

    async def _remove_reaction_listener(self):
        self._context.bot.unsubscribe(hikari.ReactionAddEvent, self._process_reaction_add)
        try:
            await self._msg.remove_all_reactions()
        except (hikari.ForbiddenError, hikari.NotFoundError):
            pass

    async def _timeout_coro(self):
        try:
            await asyncio.sleep(self._timeout)
            await self._remove_reaction_listener()
        except asyncio.CancelledError:
            pass

    async def run(self, context: Context) -> None:
        """
        Run the navigator under the given context.

        Args:
            context (:obj:`~.context.Context`): Context to run the navigator under

        Returns:
            ``None``

        Raises:
            :obj:`hikari.MissingIntentError`: If the bot does not have the relevant reaction intent(s) for
                the navigator to function.
        """
        intent_to_check_for = (
            hikari.Intents.GUILD_MESSAGE_REACTIONS
            if context.guild_id is not None
            else hikari.Intents.DM_MESSAGE_REACTIONS
        )
        if not (context.bot.intents & intent_to_check_for) == intent_to_check_for:
            raise hikari.MissingIntentError(intent_to_check_for)

        self._context = context
        context.bot.subscribe(hikari.ReactionAddEvent, self._process_reaction_add)
        self._msg = await self._send_initial_msg(self.pages[self.current_page_index])
        for emoji in [button.emoji for button in self.buttons]:
            await self._msg.add_reaction(emoji)

        if self._timeout_task is not None:
            self._timeout_task.cancel()
        self._timeout_task = asyncio.create_task(self._timeout_coro())


class StringNavigator(Navigator[str]):
    """
    A reaction navigator system for navigating through a list of string messages.

    Default buttons:

    - ``\\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\\N{VARIATION SELECTOR-16}`` (Go to first page)

    - ``\\N{BLACK LEFT-POINTING TRIANGLE}\\N{VARIATION SELECTOR-16}`` (Go to previous page)

    - ``\\N{BLACK SQUARE FOR STOP}\\N{VARIATION SELECTOR-16}`` (Stop navigation)

    - ``\\N{BLACK RIGHT-POINTING TRIANGLE}\\N{VARIATION SELECTOR-16}`` (Go to next page)

    - ``\\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}\\N{VARIATION SELECTOR-16}`` (Go to last page)

    Args:
        pages (Sequence[ :obj:`str` ]): Pages to navigate through.

    Keyword Args:
        buttons (Optional[ Sequence[ :obj:`~.utils.nav.NavButton` ] ]): Buttons to
            use the navigator with. Uses the default buttons if not specified.
        timeout (:obj:`float`): The navigator timeout in seconds. After the timeout has expired, navigator reactions
            will no longer work. Defaults to 120 (2 minutes).

    Example:

        .. code-block:: python

            from lightbulb.utils import pag, nav

            @bot.command()
            async def foo(ctx):
                paginated_help = pag.StringPaginator()
                for l in thing_that_creates_a_lot_of_text.split("\\n"):
                    paginated_help.add_line(l)
                navigator = nav.StringNavigator(paginated_help.build_pages())
                await navigator.run(ctx)

    """

    async def _edit_msg(self, message: hikari.Message, page: str) -> hikari.Message:
        return await message.edit(page)

    async def _send_initial_msg(self, page: str) -> hikari.Message:
        return await self._context.respond(page)


class EmbedNavigator(Navigator[hikari.Embed]):
    """
    A reaction navigator system for navigating through a list of embeds.

    Args:
        pages (Iterable[ :obj:`~hikari.embeds.Embed` ]): Pages to navigate through.

    Keyword Args:
        buttons (Optional[ Iterable[ :obj:`~.utils.nav.NavButton` ] ]): Buttons to
            use the navigator with. Uses the default buttons if not specified.
        timeout (:obj:`float`): The navigator timeout in seconds. After the timeout has expired, navigator reactions
            will no longer work. Defaults to 120 (2 minutes).

    Note:
        See :obj:`~.utils.nav.StringNavigator` for the default buttons supplied by the navigator.
    """

    async def _edit_msg(self, message: hikari.Message, page: hikari.Embed) -> hikari.Message:
        return await message.edit(embed=page)

    async def _send_initial_msg(self, page: hikari.Embed) -> hikari.Message:
        return await self._context.respond(embed=page)



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
"""
.. error::
    The implementation of slash commands within a Plugin class is not supported.
"""
import typing

from lightbulb.slash_commands.commands import *
from lightbulb.slash_commands.context import *

__all__: typing.Final[typing.List[str]] = [
    "SlashCommandBase",
    "TopLevelSlashCommandBase",
    "SlashCommand",
    "SlashCommandGroup",
    "SlashSubGroup",
    "SlashSubCommand",
    "SlashCommandOptionsWrapper",
    "SlashCommandContext",
]



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.Tuple[str]] = [
    "EventListenerDescriptor",
    "listener",
    "Plugin",
]

import inspect
import types
import typing

import hikari

from lightbulb import commands
from lightbulb import context as context_

T = typing.TypeVar("T")
EventT_co = typing.TypeVar("EventT_co", bound=hikari.Event, covariant=True)


class EventListenerDescriptor:
    """
    Descriptor for a listener.

    This provides the same introspective logic as :meth:`hikari.BotApp.listen`, but
    does so using a descriptor instead of directly subscribing the function. This
    is detected when loading plugins as a way of defining event listeners within
    plugins lazily.

    It may either consume an explicit event type, or introspect the given callback
    to get the type hint on the given callback for the event parameter after self.

    This will only work with instance-method style classes.
    """

    def __init__(
        self,
        event_type: typing.Optional[typing.Type[EventT_co]],
        callback: typing.Callable[[typing.Any, EventT_co], typing.Coroutine[typing.Any, typing.Any, None]],
    ) -> None:
        self.name: typing.Optional[str] = None
        self.callback = callback
        self.owner: typing.Optional[typing.Type[typing.Any]] = None

        signature = inspect.signature(callback)
        resolved_typehints = typing.get_type_hints(callback)
        params = []

        none_type = type(None)
        for name, param in signature.parameters.items():
            if isinstance(param.annotation, str):
                param = param.replace(
                    annotation=resolved_typehints[name] if name in resolved_typehints else inspect.Parameter.empty
                )
            if param.annotation is none_type:
                param = param.replace(annotation=None)
            params.append(param)

        return_annotation = resolved_typehints.get("return", inspect.Signature.empty)
        if return_annotation is none_type:
            return_annotation = None

        self.__signature__ = signature.replace(parameters=params, return_annotation=return_annotation)

        if event_type is None:
            if len(self.__signature__.parameters) != 2:
                raise TypeError(
                    f"Expected two positional parameters on event listener (self, event), "
                    f"got {len(self.__signature__.parameters)}"
                )

            param_iterator = iter(self.__signature__.parameters.values())
            next(param_iterator)  # discard "self"
            event_param = next(param_iterator)

            if event_param.kind not in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.POSITIONAL_ONLY,
            ):
                raise TypeError("Expected two positional parameters on event listener (self, event)")

            if not issubclass(event_param.annotation, hikari.Event):
                raise TypeError(
                    "Event parameter annotation type must be a subclass of hikari.Event, or an explicit "
                    "type must be given instead."
                )

            event_type = typing.cast("typing.Type[hikari.Event]", event_param.annotation)

        self.event_type = event_type

    def __set_name__(self, owner: typing.Type[typing.Any], name: str) -> None:
        self.name = name
        self.owner = owner

    @typing.no_type_check
    def __get__(self, instance: typing.Any, owner: typing.Type[typing.Any]) -> typing.Any:
        if instance is None:
            return owner
        return types.MethodType(self.callback, instance)


def listener(
    event_type: typing.Optional[typing.Type[EventT_co]] = None,
) -> typing.Callable[[T], EventListenerDescriptor]:
    """
    A decorator that registers a plugin method as an event listener.

    Args:
        event_type (Optional[ :obj:`hikari.Event` ]): The event to listen to. If
            unspecified then it will be inferred from the method's typehint.

    Example:

        .. code-block:: python

            import lightbulb
            import hikari

            class TestPlugin(lightbulb.Plugin):
                @plugins.listener(hikari.MessageCreateEvent)
                async def print_message(self, event):
                    print(event.message.content)
    """

    def decorator(listener: T) -> EventListenerDescriptor:
        return EventListenerDescriptor(event_type, listener)

    return decorator


class Plugin:
    """
    Independent class that can be loaded and unloaded from the bot
    to allow for hot-swapping of commands.

    To use in your own bot you should subclass this for each plugin
    you wish to create. Don't forget to cal ``super().__init__()`` if you
    override the ``__init__`` method.

    Args:
        name (Optional[ :obj:`str` ]): The name to register the plugin under. If unspecified will be the class name.

    Example:

        .. code-block:: python

            import lightbulb

            bot = lightbulb.Bot(...)

            class MyPlugin(lightbulb.Plugin):

                @lightbulb.command()
                async def ping(self, ctx):
                    await ctx.send("Pong!")

            bot.add_plugin(MyPlugin())
    """

    def __init__(self, *, name: str = None) -> None:
        self.name = self.__class__.__name__ if name is None else name
        """The plugin's registered name."""
        self.commands = set()
        """"A set containing all commands and groups registered to the plugin."""
        self._commands: typing.MutableMapping[str, typing.Union[commands.Command, commands.Group]] = {}
        self.listeners: typing.MutableMapping[
            typing.Type[hikari.Event],
            typing.MutableSequence[EventListenerDescriptor],
        ] = {}
        """Mapping of event to a listener method containing all listeners registered to the plugin."""

        # we use type(self) since it will prevent the descriptor __get__ being
        # invoked to convert the command to a bound instance.
        for name, member in type(self).__dict__.items():
            if isinstance(member, commands.Command):
                if not member.is_subcommand:
                    # using self here to now get the bound command.
                    self._commands[member.name] = getattr(self, name)
                    self._commands[member.name].plugin = self
                    self.commands.add(self._commands[member.name])

            elif isinstance(member, EventListenerDescriptor):
                if member.event_type not in self.listeners:
                    self.listeners[member.event_type] = []
                self.listeners[member.event_type].append(member)

    def __repr__(self) -> str:
        return f"<lightbulb.Plugin {self.name} at {hex(id(self))}>"

    def plugin_remove(self) -> None:
        """
        A method that will be called just before the plugin is removed from
        the bot. This method **cannot** be a coroutine, it must be a regular
        function.

        You may with use this for any cleanup that the plugin may require.
        """
        pass

    async def plugin_check(self, context: context_.Context) -> bool:
        """
        A check method called for only the commands inside the plugin.

        This method **must** be a coroutine and return a boolean-like value
        or raise an error when called.

        Args:
            context (:obj:`~.context.Context`): The command invocation context.

        Returns:
            :obj:`bool`: Whether the check passed or failed.
        """
        return True

    def walk_commands(self) -> typing.Generator[commands.Command, None, None]:
        """
        A generator that walks through all commands, groups and subcommands registered to this plugin.

        Yields:
            :obj:`~.commands.Command`: All commands, groups and subcommands registered to this plugin.
        """
        for command in self.commands:
            yield command
            if isinstance(command, commands.Group):
                yield from command.walk_commands()
